﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------

-- the shapeshift bar frame

nUI_ShapeshiftBar = CreateFrame( "Frame", "nUI_ShapeshiftBar", nUI_Dashboard, "SecureStateHeaderTemplate" );

nUI_ShapeshiftBar:SetBackdrop(
	{
		bgFile   = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBg.blp", 
		edgeFile = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBorder.blp", 
		tile     = true, 
		tileSize = 1, 
		edgeSize = 12, 
		insets   = {left = 0, right = 0, top = 0, bottom = 0},
	}
);

nUI_ShapeshiftBar:SetBackdropColor( 0, 0, 0, 0.35 );
nUI_ShapeshiftBar:SetAttribute( "showstates", 2 );
nUI_PetBarAnchor:SetAttribute( "addchild", nUI_ShapeshiftBar );

-- button initialization

nUI_ShapeshiftBar.Buttons = {};

for i=1,10 do

	local button    = CreateFrame( "CheckButton", "$parent_Button"..i, nUI_ShapeshiftBar, "ShapeshiftButtonTemplate" );	

	nUI_ShapeshiftBar.Buttons[i] = button;
	
	button:SetID( i );
	button:SetParent( nUI_ShapeshiftBar );
	button:SetNormalTexture( "" );
	button.SetNormalTexture = function() end;
	
	button.actionType               = "SHAPESHIFTBUTTON";
	button.actionID                 = i;
	button.nUI_CachedClearAllPoints = button.ClearAllPoints;
	button.nUI_CachedSetAllPoints   = button.SetAllPoints;
	button.nUI_CachedSetPoint       = button.SetPoint;

	button.ClearAllPoints = function() end;
	button.SetAllPoints   = function() end;
	button.SetPoint       = function() end;
	
	nUI_ShapeshiftBar:SetAttribute( "addchild", button );	
	
end

-- manage the layout of the Shapeshift Bar

nUI_ShapeshiftBar:SetScript( "OnEvent", 
	function() 
		if event == "SPELL_UPDATE_COOLDOWN"
		or event == "SPELL_UPDATE_USABLE"
		or event == "PLAYER_AURAS_CHANGED"
		or event == "UPDATE_BONUS_ACTIONBAR"
		then 
			for i=1,GetNumShapeshiftForms() do 
				nUI_ShapeshiftBar.updateState( i ); 
			end
		else 
			nUI_ShapeshiftBar:layoutFrame(); 
		end
	end 
);
	
nUI_ShapeshiftBar:RegisterEvent( "VARIABLES_LOADED" );
nUI_ShapeshiftBar:RegisterEvent( "DISPLAY_SIZE_CHANGED" );
nUI_ShapeshiftBar:RegisterEvent( "UPDATE_SHAPESHIFT_FORMS" );
nUI_ShapeshiftBar:RegisterEvent( "SPELL_UPDATE_COOLDOWN" );
nUI_ShapeshiftBar:RegisterEvent( "SPELL_UPDATE_USABLE" );
nUI_ShapeshiftBar:RegisterEvent( "PLAYER_AURAS_CHANGED" );
nUI_ShapeshiftBar:RegisterEvent( "UPDATE_BONUS_ACTIONBAR" );

nUI_ShapeshiftBar.updateState = function( index )
	
	local button = nUI_ShapeshiftBar.Buttons[index];
	
	texture, name, isActive, isCastable = GetShapeshiftFormInfo( index );
	button.layers.icon:SetTexture( texture );
		
	if ( isActive ) then button:SetChecked( 1 );
	else button:SetChecked( 0 );
	end
		
	if ( isCastable ) then button.layers.icon:SetVertexColor( 1.0, 1.0, 1.0 );
	else button.layers.icon:SetVertexColor( 0.4, 0.4, 0.4 );
	end
		
	start, duration, enable = GetShapeshiftFormCooldown( index );
	CooldownFrame_SetTimer( button.layers.cooldown, start, duration, enable );
	
end
	
nUI_ShapeshiftBar.layoutFrame = function()
				
	nUI:setScale();
				
	local btn_size = nUI_PetBarLayout.btn_size * nUI.scale;
	local btn_gap  = nUI_PetBarLayout.gap * nUI.scale;
	local width    = 0;
	local last_button;

	nUI_ShapeshiftBar:SetScale( 1.0 / nUI.scale );	
	nUI_ShapeshiftBar:SetFrameStrata( nUI_Dashboard:GetFrameStrata() );
	nUI_ShapeshiftBar:SetFrameLevel( nUI_Dashboard:GetFrameLevel()+3 );
	
	for i in pairs( nUI_ShapeshiftBar.Buttons ) do
		
		local button = nUI_ShapeshiftBar.Buttons[i];

		button:SetFrameStrata( nUI_ShapeshiftBar:GetFrameStrata() );
		button:SetFrameLevel( nUI_ShapeshiftBar:GetFrameLevel()+1 );
		button:SetScale( btn_size / button:GetHeight() );

		nUI:initActionButton( button );

		if i > GetNumShapeshiftForms() then 
		
			button:Hide();
			
		else 
		
			width = width + button:GetWidth() * button:GetEffectiveScale() + (i > 1 and btn_gap or 0);
			
			button:Show();

			button:nUI_CachedClearAllPoints();
			
			if i == 1 then button:nUI_CachedSetPoint( "TOPLEFT", nUI_ShapeshiftBar, "TOPLEFT", btn_size * 0.05, -btn_size * 0.05 );
			else button:nUI_CachedSetPoint( "TOPLEFT", last_button, "TOPRIGHT", btn_gap, 0 );
			end
			
			nUI_ShapeshiftBar.updateState( i );

		end			
		
		last_button = button;
		
	end		
	
	nUI_ShapeshiftBar:SetHeight( btn_size * 1.1 );
	nUI_ShapeshiftBar:SetWidth( width + btn_size * 0.1 );
	
	nUI_ShapeshiftBar:ClearAllPoints();
	nUI_ShapeshiftBar:SetPoint( "TOP", nUI_PetBar, "TOP", 0, 0 );
	
end